#ifndef LIB_SPRINGMODEL_H
#define LIB_SPRINGMODEL_H

#include "base/texture.h"
#include "lib/objfile.h"
#include "lib/springobj.h"

//enable this for my experimental (and non-functional) triangle-volume-preserving energy function
//#define TRIANGLE_CONSTRAINT

class SMTri {
public:
	//store the resting angles between all vertices
	//use these for torque-based springs (time to do some math!)
	float	angle[3];

#ifdef TRIANGLE_CONSTRAINT
	//resting length squared
	float	lensq[3];
#endif

	//store whether this triangle is visible or not
	//visibility is computed from a source point and later used for shadow rendering
	bool	vis;
};

/**
 * our object instance:
 */
class SpringModel : public SpringObj {
protected:

	//sized model.tri.size()
	//contains visibility information for each triangle face
	//used for shadow casting
	SMTri *tri;

	//instance model
	//source of our array size and original vertex positions
	ObjFile *model;

	virtual void clear();

public:

	Texture tex;

	//same as parent, right?
	SpringModel() {clear();}
	~SpringModel() {shutdown();}

	// unload all data
	virtual void shutdown();

	// render
	void render();

	// useful method for shadow casting.
	void renderEdges(const vec3f &offset, float pointSource, float dist) const;
	void renderEdges(const vec4f &offset, float dist) const {
		renderEdges((const vec3f&)offset, offset.w, dist);
	}

	// calculates the flags of which faces are visible
	// from the source position.  uses front/back testing for each tri's plane
	void calcVisibleFaces(const vec4f &src);

	// resets all the model vertexes to those of its parent model file
	void resetModelVertexes();

	// resets the rest length of all model edges according to the current SMVertex positions
	virtual void resetSpringLengths();

	//get/set model

	ObjFile *getModel() const { return model; }
	bool setModel(ObjFile *model);

	//physics:

	//accumulate the forces on the vertices
	//currently the only forces that the SpringModel class handles are angular springs
	//linear springs are all handled externally . . . (though they probably shouldn't be . . . )
	void applyForces();
};

#endif
