#include "math/vec.h"
#include "lib/bumpmap.h"

void BumpMap::gen(const Texture &in, Texture &out, float relative_altitude) {

	float inv_alt = 2.f / relative_altitude;

	int texwidth = in.getWidth();
	int texheight = in.getHeight();

	//gets RGB data
	unsigned char *data = (unsigned char *)in.getData();	//new'd so delete it once you're done

	//make a tangentSpace mpa
	unsigned char *tangentSpace = new unsigned char[texwidth * texheight * 4];
	typedef _vec<4, unsigned char> vec4uc;
	vec4uc *p = (vec4uc *)tangentSpace;
	for (int j = 0; j < texheight; j++) {
		for (int i = 0; i < texwidth; i++, p++) {

			int im = (i + texwidth-1) % texwidth;
			int in = (i + 1) % texwidth;
			int jm = (j + texheight-1) % texheight;
			int jn = (j + 1) % texheight;

			static vec3f greyscale(0.3f, 0.6f, 0.1f);
			float fim = dot(greyscale, (vec3f)(vec3uc&)data[3 * (im + j * texwidth)]) / 255.f;
			float fin = dot(greyscale, (vec3f)(vec3uc&)data[3 * (in + j * texwidth)]) / 255.f;
			float fjm = dot(greyscale, (vec3f)(vec3uc&)data[3 * (i + jm * texwidth)]) / 255.f;
			float fjn = dot(greyscale, (vec3f)(vec3uc&)data[3 * (i + jn * texwidth)]) / 255.f;
			//surface normal is <-dx, -dy, 2/z_scale>
			//computation of <2, 0, z * dx> x <0, 2, z * dy>
			//does ~ take precedence over + ?
			//vec3f del = ( vec3f(fim - fin, fjm - fjn, inv_alt).normalize() + 1.f) * 0.5f;
			//all in place:
			vec3f del(fim-fin, fjm-fjn, inv_alt);
			(del.normalize() += 1.f) *= 0.5f;

			*(vec3uc*)p = (vec3uc)(del * 255.f);
			//store the actual height:
			p->w = (unsigned char)( dot(greyscale, (vec3f)(vec3uc&)data[3 * (i + j * texwidth)] ) );
		}
	}
	out.build2DMipmaps(texwidth,texheight, GL_RGBA, GL_UNSIGNED_BYTE, tangentSpace);

	delete data;
	delete tangentSpace;
}