#include "lib/parsystem.h"

#define PARTICLE_MAX		3000

void ParSystem::reset() {
	live = NULL;
	dead = NULL;
	parCount = 0;
	liveCount = 0;
}

bool ParSystem::init() {
	shutdown();
	return true;
}

void ParSystem::shutdown() {

	//delete all particles

	Particle *next,*p;

	for (p = live; p; p = next) {
		next = p->next;
		delete p;
	}

	for (p = dead; p; p = next) {
		next = p->next;
		delete p;
	}

	//and clear the list pointers

	reset();
}

Particle *ParSystem::createParticle() {

	//grab or alloc a new one
	Particle *p = NULL;
	if (dead) {
		p = dead;
		dead = dead->next;
		if (dead) dead->prev = NULL;
	} else {

		parCount++;
		if (parCount > PARTICLE_MAX) return NULL;

		p = new Particle();
		if (!p) return NULL;
	}

	//link it up
	if (!live) {
		live = p;
		p->prev = NULL;
		p->next = NULL;
	} else {
		live->prev = p;
		p->next = live;
		live = p;
		p->prev = NULL;
	}

	//init our particle:
	p->clearDamage();
	p->clearForce();

	liveCount++;
	return p;
}

void ParSystem::removeParticle(Particle *p) {

	//unlink 'er first:
	if (p->prev) p->prev->next = p->next;
	else live = p->next;
	if (p->next) p->next->prev = p->prev;

	liveCount--;

	//add to the dead chain for later use
	//this prevents excessive alloc/unalloc's
	//but, unlike a vector, this doesnt leave dead particles in the same array, slowing down our search times
	//and, unlike vectors, this doesnt shift around our pointers upon resize
	p->next = dead;
	if (dead) dead->prev = p;
	dead = p;
	p->prev = NULL;
}
