#include <math.h>
#include <stdio.h>

#include "defs/defs.h"
#include "system/gl.h"
#include "lib/skymap.h"

void Skymap::shutdown() {
	skyTexCubemap.unload();
	for (int i = 0; i < 6; i++) {
		skyTexBackup[i].unload();
	}

	reset();
}

void Skymap::reset() {
	state = SKYMAP_USE_NONE;
}

//#define TEXMATFIX

//returns 2 if it can use cubemaps
//returns 1 for 2d texture maps
//returns 0 if no such load worked
skymapState_t Skymap::setSkyTex(const char *skyfn[6]) {

	shutdown();	//calls reset, which resets the state to SKYMAP_USE_NONE ...

	const char *newsky[6] = {
		skyfn[0], skyfn[1], skyfn[2], skyfn[3], skyfn[4], skyfn[5]
	};

	if (skyTexCubemap.loadCubeMap(newsky[0], newsky[1], newsky[2], newsky[3], newsky[4], newsky[5])) {
		//using cubemap
		state = SKYMAP_USE_CUBEMAP;
	} else {
		printf("unable to create sky cubemap\n");
	}

	bool gotEmAll = true;
	for (int i = 0; i < 6; i++) {
		//force RGB only so we deny any alpha channels coming thru
		if ( (gotEmAll &= skyTexBackup[i].load2D(newsky[i])) == false ) {
			printf("unable to create sky textures %s\n", skyfn[i]);
		}
	}
	//if cubemap failed and we got all our 2d textures
	//then remember to use 2d
	if (gotEmAll && state == SKYMAP_USE_NONE) {
		state = SKYMAP_USE_2D;
	}
	if (state == SKYMAP_USE_NONE) {
		printf("you might not see any skymap...\n");
	}
	
	return state;
}

void Skymap::drawSky() {

	//nothing working
	if (state == SKYMAP_USE_NONE) return;

	const int SKY_CUBE_DIST = 100;		//somewhere between view znear and zfar

	static int planes[24] = {
		6,4,5,7,
		0,2,3,1,
		2,6,7,3,
		4,0,1,5,
		1,3,7,5,
		4,6,2,0
	};
	
	static float cube_vtxs[8][3] = {
		{-SKY_CUBE_DIST,-SKY_CUBE_DIST,	-SKY_CUBE_DIST	},
		{-SKY_CUBE_DIST,-SKY_CUBE_DIST,	SKY_CUBE_DIST	},
		{-SKY_CUBE_DIST,SKY_CUBE_DIST,	-SKY_CUBE_DIST	},
		{-SKY_CUBE_DIST,SKY_CUBE_DIST,	SKY_CUBE_DIST	},
		{SKY_CUBE_DIST,	-SKY_CUBE_DIST,	-SKY_CUBE_DIST	},
		{SKY_CUBE_DIST,	-SKY_CUBE_DIST,	SKY_CUBE_DIST	},
		{SKY_CUBE_DIST,	SKY_CUBE_DIST,	-SKY_CUBE_DIST	},
		{SKY_CUBE_DIST,	SKY_CUBE_DIST,	SKY_CUBE_DIST	}

	};

	static float backup_cube_vtxs[24][2] = {
		{1,0},{1,1},{0,1},{0,0},
		{0,1},{0,0},{1,0},{1,1},
		{0,0},{1,0},{1,1},{0,1},
		{1,1},{0,1},{0,0},{1,0},
		{0,1},{0,0},{1,0},{1,1},
		{0,1},{0,0},{1,0},{1,1}
	};

	//get us the right color
	glColor3f(1,1,1);

	if (found_GL_ARB_texture_cube_map &&	//if the ext is available
		state == SKYMAP_USE_CUBEMAP)		//if the user says so
	{
		skyTexCubemap.bind();

		glBegin(GL_QUADS);
		for (int i = 0; i < 24; i++) {
			glTexCoord3fv(cube_vtxs[planes[i]]);
			glVertex3fv(cube_vtxs[planes[i]]);
		}
		glEnd();

	} else if (state == SKYMAP_USE_2D) {
		int i = 0;
		for (int f = 0; f < 6; f++) {
			skyTexBackup[f].bind();
			glBegin(GL_QUADS);
			for (int q = 0; q < 4; q++) {
				glTexCoord2fv(backup_cube_vtxs[i]);
				glVertex3fv(cube_vtxs[planes[i]]);
				i++;
			}
			glEnd();
		}
	}
}

void Skymap::render(const View *view) {

	glPushAttrib(GL_ENABLE_BIT);
//	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);

	//no touchin that depth buffer!
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	//dont worry aboutthe aspecRatio param unless we're messing with the projection
	view->setupView(bitflag(VIEW_FLAG_CLEAR_MODELVIEW) | bitflag(VIEW_FLAG_APPLY_ROTATION), 1.f);
	glRotatef(-90, 1, 0, 0);//correct for skytex

	drawSky();

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	//glClear(GL_DEPTH_BUFFER_BIT);
	glPopAttrib();
}