#include <assert.h>
#include <stdio.h>
#include "lib/springsystem.h"

void SpringSystem::reset() {
	spring.clear();
}

vec3f SpringSystem::computeSpringForce(const vec3f &dp, const vec3f &dv, float ks, float kd, float restlen) {
	float dist = (float)dp.len();

	vec3f f;
#if 1	//for now: i dont trust my lim(|dp|->0) math
	if (dist < 0.0001f) {

		return vec3f(0,0,0);

		//first calculate our 'd' component as |dp|->0
		//d = lim|dp|->0 of dv.dp*kd/|dp|
		//d = lim|dp|->0 of dv.unit_dp*|dp|*kd/|dp|
		//d = lim|dp|->0 of |dv|*unit_dv.unit_dp*kd
		//d = lim|dp|->0 of |dv|*kd*cos(theta) <- for theta being the angle between the difference in velocity and difference in position vectors
		//so how do we find the direction of dp when the magnitude is zero?  use its derivative!
		//dif[dp,t] = dv
		//d = lim|dp|->0 of |dv|*unit_dv.unit_dv*kd
		//d = |dv|*kd
		//
		//now we calculate the force:
		//f = dp * ( - (h + d) / |dp| );
		//f = unit_dp * |dp| * ( - (h + d) / |dp| );
		//f = unit_dp * (-(h + d));
		//f = unit_dv * (-(h + d))
		//and h = (dist - restlen) * ks => h = -restlen * ks
		//f = unit_dv * (-(-restlen * ks + d))
		//f = unit_dv * (restlen * ks - d)
		//substitute d = |dv| * kd
		//f = unit_dv * (restlen * ks - |dv| * kd)
		//f = unit_dv * restlen * ks - dv * kd
		f = vec3unit(dv) * (restlen * ks) - dv * kd;
	} else 
#endif
	{
		float h = (dist - restlen) * ks;
		float d = (dot(dv, dp) * kd) / dist;
		f = dp * ( - (h + d) / dist );
	}

	if (f.len() > 10000000) {
		printf("got an out of bounds force: %f %f %f\n", VEC3ELEM(f));
		f = vec3f(0,0,0);
	}
	return f;
}

void SpringSystem::applyForces(ParSystem &parSys) {
	for (int i = 0; i < (int)spring.size(); i++) {
		Spring &s = spring[i];
		if (!s.active) continue;

		Particle *p0 = s.par[0];
		Particle *p1 = s.par[1];

		//make sure both particles really are there =)
		assert(p0 && p1);

		vec3f f = computeSpringForce(p0->x - p1->x, p0->v - p1->v, s.ks, s.kd, s.restlen);
		p0->addForce(f);
		p1->addForce(-f);
	}
}

int SpringSystem::createSpring() {
	int i;
	for (i = 0; i < (int)spring.size(); i++) {
		if (!spring[i].active) {
			spring[i].active = true;
			return i;
		}
	}
	//give us some more space
	spring.resize(i + 1);
	for (int j = i; j < (int)spring.size(); j++) {
		spring[j].active = false;
	}
	spring[i].active = true;
	return i;	
}
